"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// @ts-ignore
const usfmjs = require("usfm-js");
const Reference_1 = require("./Reference");
const AlignedSentence_1 = require("./AlignedSentence");
/**
 * Injects alignments into the usfm
 * @param alignments
 * @param {string} usfm
 * @return {string}
 */
function alignUSFM3(alignments, usfm) {
    const usfmObject = usfmjs.toJSON(usfm);
    if (alignments) {
        for (const verseId of Object.keys(alignments.sentences)) {
            const sentence = AlignedSentence_1.default.fromJson(alignments.sentences[verseId]);
            const reference = Reference_1.default.buildFromContext(sentence.target.context);
            const cId = reference.chapter.toString();
            const vId = reference.verse.toString();
            // look up verse
            if (Object.keys(usfmObject.chapters).indexOf(cId) >= 0 && Object.keys(usfmObject.chapters[cId]).indexOf(vId) >= 0) {
                // apply alignments
                usfmObject.chapters[cId][vId] = alignVerse(usfmObject.chapters[cId][vId], sentence);
            }
            else {
                console.warn(`${reference} not found in usfm`);
            }
        }
    }
    return usfmjs.toUSFM(usfmObject);
}
exports.alignUSFM3 = alignUSFM3;
/**
 * Looks up an alignment index
 * @param verseObj - the verse object used for the search
 * @param verse
 * @return {number}
 */
function findAlignment(verseObj, verse) {
    for (const alignmentIndex of Object.keys(verse.alignments)) {
        const alignment = verse.alignments[parseInt(alignmentIndex)];
        for (const tokenId of alignment.targetNgram) {
            const token = verse.target.tokens[tokenId];
            if (verseObj.text == token.text
                && verseObj.occurrence == token.occurrence
                && verseObj.occurrences == token.occurrences) {
                return Number.parseInt(alignmentIndex);
            }
        }
    }
    return -1;
}
/**
 * Returns the title of the alignment source tokens
 * @param verse
 * @param {number} alignmentIndex
 * @return {string}
 */
function getAlignmentTitle(verse, alignmentIndex) {
    return verse.alignments[alignmentIndex].sourceNgram.map((i) => {
        return verse.source.tokens[i].text;
    }).join(" ");
}
/**
 * Compares two numbers for sorting
 * @param {number} a
 * @param {number} b
 * @return {number}
 */
function numberComparator(a, b) {
    return a - b;
}
/**
 * Compares two verse objects for sorting
 * @param a
 * @param b
 * @return {number}
 */
function verseObjectComparator(a, b) {
    const aSort = getVerseObjectSortKey(a);
    const bSort = getVerseObjectSortKey(b);
    return numberComparator(aSort, bSort);
}
/**
 * Returns the key used for sorting the verse object
 * @param obj
 * @return {number}
 */
function getVerseObjectSortKey(obj) {
    if (obj.children && obj.children.length) {
        return getVerseObjectSortKey(obj.children[0]);
    }
    else {
        return obj.position;
    }
}
/**
 * Removes the sorting key from the verse objects
 * @param verseObjects
 */
function cleanSortingKey(verseObjects) {
    for (const obj of verseObjects) {
        delete obj.position;
        if (obj.children) {
            cleanSortingKey(obj.children);
        }
    }
}
/**
 * Injects alignments into a verse
 * @param usfm
 * @param sentence
 */
function alignVerse(usfm, sentence) {
    const unusedObjects = [];
    const alignedObjects = {};
    const usfmTokens = [];
    for (const objIndex of Object.keys(usfm.verseObjects)) {
        const obj = usfm.verseObjects[objIndex];
        obj.position = objIndex;
        if (obj.type !== "word") {
            unusedObjects.push(obj);
            continue;
        }
        usfmTokens.push(new AlignedSentence_1.Token(obj.text, parseInt(obj.occurrence), parseInt(obj.occurrences)));
        const alignmentIndex = findAlignment(obj, sentence);
        if (alignmentIndex >= 0) {
            // add to alignment
            if (!alignedObjects[alignmentIndex]) {
                alignedObjects[alignmentIndex] = {
                    content: getAlignmentTitle(sentence, alignmentIndex),
                    tag: "zaln",
                    type: "milestone",
                    children: []
                };
            }
            alignedObjects[alignmentIndex].children.push(obj);
            alignedObjects[alignmentIndex].children.sort(verseObjectComparator);
        }
        else {
            // add to unused
            unusedObjects.push(obj);
        }
    }
    // validate tokens
    if (usfmTokens.length !== sentence.target.tokens.length) {
        throw new Error(`Sentence tokens do not match in ${sentence.target.context}`);
    }
    else {
        for (let i = 0; i < usfmTokens.length; i++) {
            if (!usfmTokens[i].equals(sentence.target.tokens[i])) {
                throw new Error(`Sentence tokens do not match in ${sentence.target.context}`);
            }
        }
    }
    const alignedUSFM = [...unusedObjects, ...Object.keys(alignedObjects).map(key => alignedObjects[key])];
    alignedUSFM.sort(verseObjectComparator);
    cleanSortingKey(alignedUSFM);
    return alignedUSFM;
}
exports.alignVerse = alignVerse;
